#include "TabuLabelPlacement.h"
using namespace Tabu;

#include <list>
#include <algorithm>
#include <fstream>
#include <cstdlib>
#include <iostream>
using namespace std;

#include <qmutex.h>
#include <qdatetime.h> 

#include "CostLabelList.h"

#define TLP TabuLabelPlacement
#define TLPM TabuLabelPlacementMap
#define TLPF TabuLabelPlacementFile

//###############//
TLP::TabuLabelPlacement()
{
	//le nombre d'itration maximum
	this->nbIterationLimit = 1000;

	//le temps d'initialisation
	this->initialisationTime = 0;

	//la taille de la liste de candidat par rapport  au nombre de label se chevauchant
	this->candidateSizeBase = 0.73;
	this->candidateReduceFactor = 1.3;
	this->candidateGrowingFactor = 15; //1.5

	//la taille de la liste tabou par rapport  au nombre de label se chevauchant
	this->tabuSize = 0.77; //0.25

	//aprs combien d'itration il faut recalculer la taille des listes (tabou et candidats)
	this->nbIterationListSizeRecompute = 47;

	//taille minimale de la liste de candidat
	this->minimalCandidateListSize = 18;

	//taille minimale de la liste tabou
	this->minimalTabuListSize = 9;
}

//###############//
TLPM::TabuLabelPlacementMap(const Map* map,
	PlacementDirectives placementDirectives, int initialPlacementSeed)
{
	this->placementDirectives = placementDirectives;

	this->map = map;
	
	this->solution = new Solution(this->map->getNbPointLabel());

	//la futur solution, par dfaut on les places  la meilleure position
	this->solutionTabu = new SolutionTabu(this->map->getNbPointLabel());

	//cre la solution initiale
	this->creatInitialSolution(this->map->getNbPointLabel(),
		this->placementDirectives.getRatioOverlap(), initialPlacementSeed);

	//le gestionnaire de chevauchement
	this->om = new OverlapManager(this->map, this->solutionTabu,
		this->placementDirectives);

	this->initialisationTime = this->om->getInitTime();

	//convertit la solution tabu en solution gnrique
	this->convSolution();
}

//###############//
TLPF::TabuLabelPlacementFile(char* filename,
	PlacementDirectives placementDirectives, int initialPlacementSeed)
{
	this->placementDirectives = placementDirectives;

	ifstream file(filename);	
	if(!file)
		this->fileError(filename);

	int nbLabel, nbPosition;

	file >> nbLabel;
	file >> nbPosition;

	//solution qui ne servira  rien puisque aucune coordonn
	//n'est fournit dans les fichiers de donnes
	this->solution = new Solution(nbLabel);

	//la futur solution, par dfaut on les places  la meilleure position
	this->solutionTabu = new SolutionTabu(nbLabel);

	//cre la solution initiale
	this->creatInitialSolution(nbLabel,
		this->placementDirectives.getRatioOverlap(), initialPlacementSeed);

	this->placementDirectives.setNbPosition(nbPosition);

	//le gestionnaire de chevauchement
	this->om = new OverlapManager(file, this->solutionTabu,
		this->placementDirectives);

	this->initialisationTime = this->om->getInitTime();

	//convertit la solution tabu en solution gnrique
	this->convSolution();
}

//###############//
void TLP::compute()
{
	//pour mesurer le temps
	QTime time;
	time.start();

	double candidateSize = this->candidateSizeBase;

	const int INFINITE = 1000000000;

	//la valeur de la meilleure solution
	int bestSolutionValue = INFINITE;

	//la liste des cots
	CostLabelList cll(*this->solutionTabu, *this->om);

	//les candidats
	CostLabelList::iterator candidates;
	int candidateListLength; //la taille de la liste des candidats

	this->nbIteration = 0; //compte le nombre d'itration

	//la liste des label tabous
	list<int> tabuList;
	int tabuListLength = 0; //la taille de la liste tabou

	//le nombre initial de chevauchement (pour information)
	nbInitialOverlap = this->om->numberOfLabelThatOverlap();

	//le nombre de chevauchement
	int numberOfLabelThatOverlap = nbInitialOverlap;

	//tant qu'il y a encore des chevauchement et que le nombre d'itration n'est pas atteint
	while(this->nbIteration != nbIterationLimit && numberOfLabelThatOverlap != 0)
	{
		//le facteur de la taille de la liste de candidat diminue
		//si elle est plus lev que le facteur de base
		if(candidateSize > candidateSizeBase) candidateSize /= candidateReduceFactor;

		//toute les n itrations : recalul la taille de la liste de candidat et tabou
		if(this->nbIteration % nbIterationListSizeRecompute == 0)
		{	
			tabuListLength = minimalTabuListSize + int(this->tabuSize * numberOfLabelThatOverlap);
			
			candidateListLength = minimalCandidateListSize +
				int(candidateSize * numberOfLabelThatOverlap);

			//si la liste des candidats excde le nombre de label
			if(candidateListLength > int(this->solutionTabu->size()))
				candidateListLength = int(this->solutionTabu->size());
		}

		candidates = cll.end() - candidateListLength;

		//cherche  modifier l'emplacement d'un label de la liste
		//des candidats afin de trouver le cot le plus faible
		double bestCostDif = -INFINITE;
		double bestCost = 0;
		CostLabelList::iterator labelListCost;
		int bestPosition = -1;

		for(CostLabelList::iterator i = candidates; i != cll.end(); i++)
		{
			//pour chaque position possible
			for(int pos = 0; pos<this->placementDirectives.getNbPosition(); pos++)
			{
				if(pos == (*this->solutionTabu)[i->numLabel]) continue;

				double cost = this->om->cost(i->numLabel, pos);

				//si le cot est meilleur que le meilleur trouv et qu'il ne se trouve pas
				//dans la liste tabou ou si il amliore la solution actuelle
				//alors le mmorise comme meilleur cot courant
				if((i->cost - cost > bestCostDif) &&
					(i->cost - cost > 0 ||
					find(tabuList.begin(), tabuList.end(), i->numLabel) == tabuList.end()))
				{
					bestCostDif = i->cost - cost;
					bestCost = cost;
					labelListCost = i;
					bestPosition = pos;
				}
			}
 		}

		this->nbIteration++;

		//si parmis les candidats aucun n'a p tre choisit 
		//(ils font tous partit de la liste tabou) alors on augmente temporairement
		//la taille de la liste de candidats
		if(bestPosition == -1)
		{
			//on limite la taille pour viter la dgnration
			if(candidateSize < 10e4) candidateSize *= candidateGrowingFactor;
			
			candidateListLength = minimalCandidateListSize +
				int(candidateSize * numberOfLabelThatOverlap);
			
			if(candidateListLength >  int(this->solutionTabu->size()))
				candidateListLength =  int(this->solutionTabu->size());
			continue;
		};

		//ajoute le label  la liste tabou
		if(find(tabuList.begin(), tabuList.end(), labelListCost->numLabel) == tabuList.end())
			tabuList.push_front(labelListCost->numLabel);
		if(int(tabuList.size()) > tabuListLength) tabuList.resize(tabuListLength);

		//met  jour la position
		(*this->solutionTabu)[labelListCost->numLabel] = bestPosition;

		//met  jour le cout du label
		cll.updateCost(labelListCost, bestCost);

		numberOfLabelThatOverlap = this->om->numberOfLabelThatOverlap();
		if(numberOfLabelThatOverlap < bestSolutionValue)
			bestSolutionValue = numberOfLabelThatOverlap;

		//mode tape par tape
		if(this->stepByStep)
		{
			QTime time2;
			time2.start();
			this->solution->setNumberOfOverlap(numberOfLabelThatOverlap);
			this->convSolution();
			this->computationTime = time.elapsed();
			this->stepByStepMutex.lock();
			time = time.addMSecs(time2.elapsed());
		}
	}

	this->solution->setNumberOfOverlap(bestSolutionValue);

	this->computationTime = time.elapsed();

	//convertit la solution tabu en solution gnrique
	this->convSolution();
}

//###############//  
void TLPM::hideLabels()
{
	this->om->hideLabels(this->map, this->solution);
}

//###############//  
void TLPM::convSolution()
{
	double length = this->placementDirectives.getPointLabelLength();

	//parcours tous les labels
	for(int i=0; i<this->map->getNbPointLabel(); i++)
		this->solution->setLabelPlacement(i,
			(*this->solutionTabu)[i] == 0 ?
				Point<double>(length, length) :
			(*this->solutionTabu)[i] == 1 ?
				Point<double>(-length - this->map->getPointLabel(i)->l, length) :
			(*this->solutionTabu)[i] == 2 ?
				Point<double>(-length - this->map->getPointLabel(i)->l,
					-length - this->map->getPointLabel(i)->h) :
				Point<double>(length,
					-length  - this->map->getPointLabel(i)->h),
			this->om->overlap(i, false, -1) > 0
		);
}

//###############//  
void TLPF::convSolution()
{
	//parcours tous les labels
	for(unsigned int i=0; i<this->solutionTabu->size(); i++)
		this->solution->setLabelPlacement(i, Point<double>(0,0), false); 		
}

//###############//  
void TLP::creatInitialSolution(int nbLabel, double ratioOverlap, int initialPlacementSeed)
{
	int bestPosition = 0;

	//si la position des labels n'a pas d'importance alors
	//on va les placer initialement de manire alatoire
	if(ratioOverlap != 1.0)
	{
		//cherche la meilleurs position pour le placement initial des labels
		double bestPositionValue = 99;
		for(int i=0; i<this->placementDirectives.getNbPosition(); i++)
			if(this->placementDirectives.getPreference(i) < bestPositionValue)
			{
				bestPositionValue =this->placementDirectives.getPreference(i);
				bestPosition = i;
			}
	}
	else
	{
		srand(initialPlacementSeed == -1 ? time(0) : initialPlacementSeed);
	}	

	for(int i=0; i<nbLabel; i++)
		(*this->solutionTabu)[i] = ratioOverlap != 1.0 ? bestPosition :
			int((double(rand())/RAND_MAX)*4);
}

//###############//  
void TLP::fileError(char* name)
{
	cerr << "Error : the file " << name << " doesn't exist" << endl;
	exit(1);
}

#undef TLP
#undef TLPM
#undef TLPF
