/**
  * Classes effectuant l'optimisation de placement de label
  *  l'aide de l'heuristique 'Tabou'.
  */

#ifndef _TABU_TABULABELPLACEMENT_H_
#define _TABU_TABULABELPLACEMENT_H_

//les directives de placements
#include "PlacementDirectives.h"
using namespace Tabu;

#include <vector>
using namespace std;

//la structure de donnes reprsentant les labels  placer
#include "../Map.h"

//la structure de donnes repsentant la solution du placement
#include "../Solution.h"
#include "../LabelPlacement.h"

#include "SolutionTabu.h"
#include "OverlapManager.h"

namespace Tabu
{
	class TabuLabelPlacement : public LabelPlacement
	{
	public :

		/**
		  * Constructeur,
		  * Initialise certains paramtres par dfaut.
		  */
		TabuLabelPlacement();

		/**
		  * Permet de dfinir les variables de l'algorithme.
		  * p1 : le nombre d'itration maximum
		  * p2 : le facteur dfinissant la taille de la liste de cadidats de base
		  * p3 : le facteur de rduction du facteur p3
		  * p4 : le facteur d'accroissement du facteur p3
		  * p5 : la taille de la liste tabou en fonction du nombre de chevauchement
		  * p6 : le nombre d'itration avant de recalculer les tailles des listes
		  * (candidat et tabou)
		  * p7 : la taille minimale de la liste des candidats
		  * p8 : la taille minimal de la liste tabou 
		  */
		void setVariables(int nbIterationLimit,
			double candidateSizeBase, double candidateReduceFactor,
			double candidateGrowingFactor,
			double tabuSize, int nbIterationListSizeRecompute,
			int minimalCandidateListSize, int minimalTabuListSize)
		{
			this->nbIterationLimit = nbIterationLimit;

			this->candidateSizeBase = candidateSizeBase;
			this->candidateReduceFactor = candidateReduceFactor;
			this->candidateGrowingFactor = candidateGrowingFactor;
			this->tabuSize = tabuSize;
			this->nbIterationListSizeRecompute = nbIterationListSizeRecompute;
			this->minimalCandidateListSize = minimalCandidateListSize;
			this->minimalTabuListSize = minimalTabuListSize;
		}

		virtual ~TabuLabelPlacement()
		{
			delete this->om;
			delete this->solutionTabu;
		}

		/**
		  * Renvois le nombre d'itrations.
		  */
		int getNbIteration() const
		{
			return this->nbIteration;
		}

		/**
		  * Renvois le nombre de chevauchement initial
		  */
		int getNbInitialOverlap() const
		{
			return this->nbInitialOverlap;
		}

		/**
		  * Renvois le nombre de label (point).
		  */
		int getNumberPoint() const
		{
			return int(this->solutionTabu->size());
		}

		/**
		  * Retourne le temps d'initialisation.
		  */
		int getInitialisationTime() const
		{
			return this->initialisationTime;
		}

		/**
		  * Commence le calcul.
		  */
		virtual void compute();

	protected :

		/**
		  * Convertit une "solution tabou" en une solution gnrique.
		  */
		virtual void convSolution()=0;

		/**
		  * Cre la solution de dpart qui consiste  placer les labels dans la position prfre.
		  * (mme position pour tous les label)
		  * p1 : le nombre de label
		  * p2 : l'importance du chevauchement par rapport  la position
		  * p3 : le germe de gnration de la solution initiale
		  */
		virtual void creatInitialSolution(int, double, int);

		/**
		  * Est appel lorsqu'un fichier de donnes n'est pas trouv.
		  * p1 : le nom du fichier
		  */
		void fileError(char*);

		//le gestionnaire de placement 
		OverlapManager* om;

		//les directives de placement (prfrences de la position des labels)
		PlacementDirectives placementDirectives;

		//la solution : un tableau d'entier ou chaque nombre indique le placement du label
		SolutionTabu* solutionTabu;

		int nbIteration; //le nombre d'itration

		int initialisationTime; //le temp d'initialisation [ms]

		//le nombre de chevauchement initial
		int nbInitialOverlap;

		//le nombre d'itration maximum
		int nbIterationLimit;

		//la taille de la liste de candidat par rapport  au nombre de label se chevauchant
		double candidateSizeBase;
		double candidateReduceFactor;
		double candidateGrowingFactor;

		//la taille de la liste tabou par rapport  au nombre de label se chevauchant
		double tabuSize;

		//aprs combien d'itration il faut recalculer la taille des listes (tabou et candidats)
		int nbIterationListSizeRecompute;

		//taille minimale de la liste de candidat
		int minimalCandidateListSize;

		//taille minimale de la liste tabou
		int minimalTabuListSize;
	};

	class TabuLabelPlacementMap : virtual public TabuLabelPlacement
	{
		public :

		/**
		  * Constructeur.
		  * p1 : les donnes : les points ou il faut mettre un label
		  * p2 : les directive de placement, par exemple les prfrences de placement 
		  * p3 : le germe de la configuration initiale
		  * d'un label autour d'un point.
		  */
		TabuLabelPlacementMap(const Map*, PlacementDirectives, int initialPlacementSeed=-1);
		TabuLabelPlacementMap(){}

		/**
		  * Dfinit les labels  masquer en fonction de leur poinds
		  */
		void hideLabels();

		protected :

		/**
		  * Convertit une "solution tabou" en une solution gnrique.
		  */
		void convSolution();

		const Map* map;
	};

	class TabuLabelPlacementFile : virtual public TabuLabelPlacement
	{
		public :

		/**
		  * Constructeur.
		  * p1 : le fichier contenant les donnes
		  * p2 : les directive de placement, par exemple les prfrences de placement 
		  * p3 : le germe de la configuration initiale
		  * d'un label autour d'un point.
		  */
		TabuLabelPlacementFile(char*, PlacementDirectives, int initialPlacementSeed=-1);
		TabuLabelPlacementFile(){}

		protected :

		/**
		  * Convertit une "solution tabou" en une solution gnrique.
		  */
		void convSolution();
	};
}

#endif
